/*
 * kgraft_patch_bsc1090869
 *
 * Fix for CVE-2018-1087, bsc#1090869
 *
 *  Upstream commit:
 *  32d43cd391ba ("kvm/x86: fix icebp instruction handling")
 *
 *  SLE12 commit:
 *  4b5624e79ce784cd07ae0873bd6854e0a275af49
 *
 *  SLE12-SP1 commit
 *  106954c4a4bc28264c7dbdb628c42a8bfb18cb4c
 *
 *  SLE12-SP2 commit:
 *  8436846d3bc9c9313aae41da2679cdbe035dd154
 *
 *  SLE12-SP3 commit:
 *  95d3e6c3bd7ee22b1bf33b4e4f91eedd3dc26f0a (stable 4.4.125)
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_KVM_INTEL)

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/list.h>
#include <linux/ktime.h>
#include <linux/hrtimer.h>
#include <linux/tracepoint.h>
#include <linux/kvm_host.h>
#include <linux/kvm_types.h>
#include <asm/vmx.h>
#include <uapi/asm/kvm.h>
#include "kgr_patch_bsc1090869.h"

#if !IS_MODULE(CONFIG_KVM_INTEL)
#error "KGR patch supports only CONFIG_KVM_INTEL=m."
#endif

#if !IS_MODULE(CONFIG_KVM)
#error "KGR patch supports only CONFIG_KVM=m."
#endif


#define KGR_PATCHED_MODULE "kvm_intel"


#define __KGR_DECLARE_TRACE(name, proto, args, cond, data_proto, data_args)	\
	static struct tracepoint *kgr__tracepoint_##name;		\
	static inline void kgr_trace_##name(proto)			\
	{								\
		if (unlikely(static_key_enabled(&kgr__tracepoint_##name->key))) \
			__DO_TRACE(kgr__tracepoint_##name,		\
				TP_PROTO(data_proto),			\
				TP_ARGS(data_args),			\
				TP_CONDITION(cond),,);			\
		if (IS_ENABLED(CONFIG_LOCKDEP) && (cond)) {		\
			rcu_read_lock_sched_notrace();			\
			rcu_dereference_sched(kgr__tracepoint_##name->funcs);	\
			rcu_read_unlock_sched_notrace();		\
		}							\
	}								\

#define KGR_DECLARE_TRACE(name, proto, args)				\
	__KGR_DECLARE_TRACE(name, PARAMS(proto), PARAMS(args),		\
			cpu_online(raw_smp_processor_id()),		\
			PARAMS(void *__data, proto),			\
			PARAMS(__data, args))

#define KGR_TRACE_EVENT(name, proto, args)			\
	KGR_DECLARE_TRACE(name, PARAMS(proto), PARAMS(args))


/* see arch/x86/kvm/trace.h */
KGR_TRACE_EVENT(kvm_page_fault,
	TP_PROTO(unsigned long fault_address, unsigned int error_code),
	TP_ARGS(fault_address, error_code)
);

static struct kvm_x86_ops **kgr_kvm_x86_ops;
static bool *kgr_kvm_rebooting;
static asmlinkage void (*kgr_kvm_spurious_fault)(void);
static void (*kgr_kvm_queue_exception)(struct kvm_vcpu *vcpu, unsigned nr);
static void (*kgr_kvm_queue_exception_e)(struct kvm_vcpu *vcpu, unsigned nr,
					 u32 error_code);
static int (*kgr_x86_emulate_instruction)(struct kvm_vcpu *vcpu,
					  unsigned long cr2,
					  int emulation_type,
					  void *insn,
					  int insn_len);
static int (*kgr_kvm_mmu_unprotect_page_virt)(struct kvm_vcpu *vcpu, gva_t gva);
static int (*kgr_kvm_mmu_page_fault)(struct kvm_vcpu *vcpu, gva_t cr2,
				     u32 error_code, void *insn, int insn_len);
static int (*kgr_kvm_vcpu_halt)(struct kvm_vcpu *vcpu);

static bool *kgr_enable_ept;
static int (*kgr_handle_machine_check)(struct kvm_vcpu *vcpu);
static void (*kgr_vmx_fpu_activate)(struct kvm_vcpu *vcpu);
static void (*kgr_skip_emulated_instruction)(struct kvm_vcpu *vcpu);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "kvm:__tracepoint_kvm_page_fault",
			(void *)&kgr__tracepoint_kvm_page_fault },
	{ "kvm:kvm_x86_ops", (void *)&kgr_kvm_x86_ops },
	{ "kvm:kvm_rebooting", (void *)&kgr_kvm_rebooting },
	{ "kvm:kvm_spurious_fault", (void *)&kgr_kvm_spurious_fault},
	{ "kvm:kvm_queue_exception", (void *)&kgr_kvm_queue_exception },
	{ "kvm:kvm_queue_exception_e", (void *)&kgr_kvm_queue_exception_e },
	{ "kvm:x86_emulate_instruction", (void *)&kgr_x86_emulate_instruction },
	{ "kvm:kvm_mmu_unprotect_page_virt",
			(void *)&kgr_kvm_mmu_unprotect_page_virt },
	{ "kvm:kvm_mmu_page_fault", (void *)&kgr_kvm_mmu_page_fault },
	{ "kvm:kvm_vcpu_halt", (void *)&kgr_kvm_vcpu_halt },
	{ "kvm_intel:enable_ept", (void *)&kgr_enable_ept },
	{ "kvm_intel:handle_machine_check", (void *)&kgr_handle_machine_check },
	{ "kvm_intel:vmx_fpu_activate", (void *)&kgr_vmx_fpu_activate },
	{ "kvm_intel:skip_emulated_instruction",
			(void *)&kgr_skip_emulated_instruction },
};


/* from arch/x86/include/asm/kvm_host.h */
/* avoid module dependency on kvm.ko */
static inline int kgr_emulate_instruction(struct kvm_vcpu *vcpu,
					  int emulation_type)
{
	return kgr_x86_emulate_instruction(vcpu, 0, emulation_type, NULL, 0);
}

/*
 * Avoid module dependency on kvm.ko.
 *
 * This is the original macro but with the load from kvm_rebooting
 * replaced by an indirect load from *kgr_kvm_rebooting and the call
 * to kvm_spurious_fault() replaced by an (retpoline-like) indirect
 * call to *kgr_kvm_spurious_fault.
 */
#define kgr____kvm_handle_fault_on_reboot(insn, cleanup_insn)	\
	"666: " insn "\n\t"				\
	"668: \n\t"					\
	".pushsection .fixup, \"ax\" \n"		\
	"667: \n\t"					\
	cleanup_insn "\n\t"				\
	"pushq %%rax \n\t"				\
	"movq kgr_kvm_rebooting, %%rax\n\t"		\
	"cmpb $0, (%%rax) \n\t"				\
	"popq %%rax \n\t"				\
	"jne 668b \n\t"				\
	__ASM_SIZE(push) " $666b \n\t"			\
	"movq kgr_kvm_spurious_fault, %%rax \n\t"	\
	"call *%%rax\n\t"				\
	".popsection \n\t"				\
	_ASM_EXTABLE(666b, 667b)


/* from arch/x86/kvm/kvm_cache_regs.h */
/* inlined */
static inline unsigned long kgr_kvm_register_read(struct kvm_vcpu *vcpu,
						  enum kvm_reg reg)
{
	if (!test_bit(reg, (unsigned long *)&vcpu->arch.regs_avail))
		(*kgr_kvm_x86_ops)->cache_reg(vcpu, reg);

	return vcpu->arch.regs[reg];
}

/* inlined */
static inline unsigned long kgr_kvm_rip_read(struct kvm_vcpu *vcpu)
{
	return kgr_kvm_register_read(vcpu, VCPU_REGS_RIP);
}

/* inlined */
static inline bool kgr_is_guest_mode(struct kvm_vcpu *vcpu)
{
	return vcpu->arch.hflags & HF_GUEST_MASK;
}


/* from arch/x86/kvm/x86.h */
/* inlined */
static inline bool kgr_kvm_event_needs_reinjection(struct kvm_vcpu *vcpu)
{
	return vcpu->arch.exception.pending || vcpu->arch.interrupt.pending ||
		vcpu->arch.nmi_injected;
}


/* from arch/x86/kvm/vmx.c */
#define kgr__ex_clear(x, reg) \
	kgr____kvm_handle_fault_on_reboot(x, "xor " reg " , " reg)

#define NR_AUTOLOAD_MSRS 8

struct loaded_vmcs {
	struct vmcs *vmcs;
	int cpu;
	int launched;
	struct list_head loaded_vmcss_on_cpu_link;
};

struct nested_vmx {
	/* Has the level1 guest done vmxon? */
	bool vmxon;
	gpa_t vmxon_ptr;

	/* The guest-physical address of the current VMCS L1 keeps for L2 */
	gpa_t current_vmptr;
	/* The host-usable pointer to the above */
	struct page *current_vmcs12_page;
	struct vmcs12 *current_vmcs12;
	struct vmcs *current_shadow_vmcs;
	/*
	 * Indicates if the shadow vmcs must be updated with the
	 * data hold by vmcs12
	 */
	bool sync_shadow_vmcs;

	/* vmcs02_list cache of VMCSs recently used to run L2 guests */
	struct list_head vmcs02_pool;
	int vmcs02_num;
	u64 vmcs01_tsc_offset;
	bool change_vmcs01_virtual_x2apic_mode;
	/* L2 must run next, and mustn't decide to exit to L1. */
	bool nested_run_pending;
	/*
	 * Guest pages referred to in vmcs02 with host-physical pointers, so
	 * we must keep them pinned while L2 runs.
	 */
	struct page *apic_access_page;
	struct page *virtual_apic_page;
	struct page *pi_desc_page;
	struct pi_desc *pi_desc;
	bool pi_pending;
	u16 posted_intr_nv;

	unsigned long *msr_bitmap;

	struct hrtimer preemption_timer;
	bool preemption_timer_expired;

	/* to migrate it to L2 if VM_ENTRY_LOAD_DEBUG_CONTROLS is off */
	u64 vmcs01_debugctl;

	u16 vpid02;
	u16 last_vpid;

	u32 nested_vmx_procbased_ctls_low;
	u32 nested_vmx_procbased_ctls_high;
	u32 nested_vmx_true_procbased_ctls_low;
	u32 nested_vmx_secondary_ctls_low;
	u32 nested_vmx_secondary_ctls_high;
	u32 nested_vmx_pinbased_ctls_low;
	u32 nested_vmx_pinbased_ctls_high;
	u32 nested_vmx_exit_ctls_low;
	u32 nested_vmx_exit_ctls_high;
	u32 nested_vmx_true_exit_ctls_low;
	u32 nested_vmx_entry_ctls_low;
	u32 nested_vmx_entry_ctls_high;
	u32 nested_vmx_true_entry_ctls_low;
	u32 nested_vmx_misc_low;
	u32 nested_vmx_misc_high;
	u32 nested_vmx_ept_caps;
	u32 nested_vmx_vpid_caps;
};

struct pi_desc {
	u32 pir[8];     /* Posted interrupt requested */
	union {
		struct {
				/* bit 256 - Outstanding Notification */
			u16	on	: 1,
				/* bit 257 - Suppress Notification */
				sn	: 1,
				/* bit 271:258 - Reserved */
				rsvd_1	: 14;
				/* bit 279:272 - Notification Vector */
			u8	nv;
				/* bit 287:280 - Reserved */
			u8	rsvd_2;
				/* bit 319:288 - Notification Destination */
			u32	ndst;
		};
		u64 control;
	};
	u32 rsvd[6];
} __aligned(64);

struct vcpu_vmx {
	struct kvm_vcpu       vcpu;
	unsigned long         host_rsp;
	u8                    fail;
	bool                  nmi_known_unmasked;
	u32                   exit_intr_info;
	u32                   idt_vectoring_info;
	ulong                 rflags;
	struct shared_msr_entry *guest_msrs;
	int                   nmsrs;
	int                   save_nmsrs;
	unsigned long	      host_idt_base;
#ifdef CONFIG_X86_64
	u64 		      msr_host_kernel_gs_base;
	u64 		      msr_guest_kernel_gs_base;
#endif
	u32 vm_entry_controls_shadow;
	u32 vm_exit_controls_shadow;
	/*
	 * loaded_vmcs points to the VMCS currently used in this vcpu. For a
	 * non-nested (L1) guest, it always points to vmcs01. For a nested
	 * guest (L2), it points to a different VMCS.
	 */
	struct loaded_vmcs    vmcs01;
	struct loaded_vmcs   *loaded_vmcs;
	bool                  __launched; /* temporary, used in vmx_vcpu_run */
	struct msr_autoload {
		unsigned nr;
		struct vmx_msr_entry guest[NR_AUTOLOAD_MSRS];
		struct vmx_msr_entry host[NR_AUTOLOAD_MSRS];
	} msr_autoload;
	struct {
		int           loaded;
		u16           fs_sel, gs_sel, ldt_sel;
#ifdef CONFIG_X86_64
		u16           ds_sel, es_sel;
#endif
		int           gs_ldt_reload_needed;
		int           fs_reload_needed;
		u64           msr_host_bndcfgs;
		unsigned long vmcs_host_cr4;	/* May not match real cr4 */
	} host_state;
	struct {
		int vm86_active;
		ulong save_rflags;
		struct kvm_segment segs[8];
	} rmode;
	struct {
		u32 bitmask; /* 4 bits per segment (1 bit per field) */
		struct kvm_save_segment {
			u16 selector;
			unsigned long base;
			u32 limit;
			u32 ar;
		} seg[8];
	} segment_cache;
	int vpid;
	bool emulation_required;

	/* Support for vnmi-less CPUs */
	int soft_vnmi_blocked;
	ktime_t entry_time;
	s64 vnmi_blocked_time;
	u32 exit_reason;

	/* Posted interrupt descriptor */
	struct pi_desc pi_desc;

	/* Support for a guest hypervisor (nested VMX) */
	struct nested_vmx nested;

	/* Dynamic PLE window. */
	int ple_window;
	bool ple_window_dirty;

	/* Support for PML */
#define PML_ENTITY_NUM		512
	struct page *pml_pg;

	u64 current_tsc_ratio;

	u32 guest_pkru;
	u32 host_pkru;

	/*
	 * Only bits masked by msr_ia32_feature_control_valid_bits can be set in
	 * msr_ia32_feature_control. FEATURE_CONTROL_LOCKED is always included
	 * in msr_ia32_feature_control_valid_bits.
	 */
	u64 msr_ia32_feature_control;
	u64 msr_ia32_feature_control_valid_bits;
};

/* inlined */
static inline struct vcpu_vmx *kgr_to_vmx(struct kvm_vcpu *vcpu)
{
	return container_of(vcpu, struct vcpu_vmx, vcpu);
}

/* inlined */
static inline bool kgr_is_page_fault(u32 intr_info)
{
	return (intr_info & (INTR_INFO_INTR_TYPE_MASK | INTR_INFO_VECTOR_MASK |
			     INTR_INFO_VALID_MASK)) ==
		(INTR_TYPE_HARD_EXCEPTION | PF_VECTOR | INTR_INFO_VALID_MASK);
}

/* inlined */
static inline bool kgr_is_no_device(u32 intr_info)
{
	return (intr_info & (INTR_INFO_INTR_TYPE_MASK | INTR_INFO_VECTOR_MASK |
			     INTR_INFO_VALID_MASK)) ==
		(INTR_TYPE_HARD_EXCEPTION | NM_VECTOR | INTR_INFO_VALID_MASK);
}

/* inlined */
static inline bool kgr_is_invalid_opcode(u32 intr_info)
{
	return (intr_info & (INTR_INFO_INTR_TYPE_MASK | INTR_INFO_VECTOR_MASK |
			     INTR_INFO_VALID_MASK)) ==
		(INTR_TYPE_HARD_EXCEPTION | UD_VECTOR | INTR_INFO_VALID_MASK);
}

/* inlined */
static inline bool kgr_is_machine_check(u32 intr_info)
{
	return (intr_info & (INTR_INFO_INTR_TYPE_MASK | INTR_INFO_VECTOR_MASK |
			     INTR_INFO_VALID_MASK)) ==
		(INTR_TYPE_HARD_EXCEPTION | MC_VECTOR | INTR_INFO_VALID_MASK);
}

/* inlined */
static inline bool kgr_is_nmi(u32 intr_info)
{
	return (intr_info & (INTR_INFO_INTR_TYPE_MASK | INTR_INFO_VALID_MASK))
		== (INTR_TYPE_NMI_INTR | INTR_INFO_VALID_MASK);
}

/* inlined */
static __always_inline unsigned long kgr_vmcs_readl(unsigned long field)
{
	unsigned long value;

	asm volatile (kgr__ex_clear(ASM_VMX_VMREAD_RDX_RAX, "%0")
		      : "=a"(value) : "d"(field) : "cc");
	return value;
}

/* inlined */
static __always_inline u32 kgr_vmcs_read32(unsigned long field)
{
	return kgr_vmcs_readl(field);
}

/* inlined */
static bool kgr_rmode_exception(struct kvm_vcpu *vcpu, int vec)
{
	switch (vec) {
	case BP_VECTOR:
		/*
		 * Update instruction length as we may reinject the exception
		 * from user space while in guest debugging mode.
		 */
		kgr_to_vmx(vcpu)->vcpu.arch.event_exit_inst_len =
			kgr_vmcs_read32(VM_EXIT_INSTRUCTION_LEN);
		if (vcpu->guest_debug & KVM_GUESTDBG_USE_SW_BP)
			return false;
		/* fall through */
	case DB_VECTOR:
		if (vcpu->guest_debug &
			(KVM_GUESTDBG_SINGLESTEP | KVM_GUESTDBG_USE_HW_BP))
			return false;
		/* fall through */
	case DE_VECTOR:
	case OF_VECTOR:
	case BR_VECTOR:
	case UD_VECTOR:
	case DF_VECTOR:
	case SS_VECTOR:
	case GP_VECTOR:
	case MF_VECTOR:
		return true;
	break;
	}
	return false;
}

/* inlined */
static int kgr_handle_rmode_exception(struct kvm_vcpu *vcpu,
				      int vec, u32 err_code)
{
	/*
	 * Instruction with address size override prefix opcode 0x67
	 * Cause the #SS fault with 0 error code in VM86 mode.
	 */
	if (((vec == GP_VECTOR) || (vec == SS_VECTOR)) && err_code == 0) {
		if (kgr_emulate_instruction(vcpu, 0) == EMULATE_DONE) {
			if (vcpu->arch.halt_request) {
				vcpu->arch.halt_request = 0;
				return kgr_kvm_vcpu_halt(vcpu);
			}
			return 1;
		}
		return 0;
	}

	/*
	 * Forward all other exceptions that are valid in real mode.
	 * FIXME: Breaks guest debugging in real mode, needs to be fixed with
	 *        the required debugging infrastructure rework.
	 */
	kgr_kvm_queue_exception(vcpu, vec);
	return 1;
}



/* new */
#define KGR_INTR_TYPE_PRIV_SW_EXCEPTION    (5 << 8) /* ICE breakpoint - undocumented */

/* new */
static inline bool kgr_is_icebp(u32 intr_info)
{
	return (intr_info & (INTR_INFO_INTR_TYPE_MASK | INTR_INFO_VALID_MASK))
		== (KGR_INTR_TYPE_PRIV_SW_EXCEPTION | INTR_INFO_VALID_MASK);
}

/* patched */
int kgr_handle_exception(struct kvm_vcpu *vcpu)
{
	struct vcpu_vmx *vmx = kgr_to_vmx(vcpu);
	struct kvm_run *kvm_run = vcpu->run;
	u32 intr_info, ex_no, error_code;
	unsigned long cr2, rip, dr6;
	u32 vect_info;
	enum emulation_result er;

	vect_info = vmx->idt_vectoring_info;
	intr_info = vmx->exit_intr_info;

	if (kgr_is_machine_check(intr_info))
		return kgr_handle_machine_check(vcpu);

	if (kgr_is_nmi(intr_info))
		return 1;  /* already handled by vmx_vcpu_run() */

	if (kgr_is_no_device(intr_info)) {
		kgr_vmx_fpu_activate(vcpu);
		return 1;
	}

	if (kgr_is_invalid_opcode(intr_info)) {
		if (kgr_is_guest_mode(vcpu)) {
			kgr_kvm_queue_exception(vcpu, UD_VECTOR);
			return 1;
		}
		er = kgr_emulate_instruction(vcpu, EMULTYPE_TRAP_UD);
		if (er != EMULATE_DONE)
			kgr_kvm_queue_exception(vcpu, UD_VECTOR);
		return 1;
	}

	error_code = 0;
	if (intr_info & INTR_INFO_DELIVER_CODE_MASK)
		error_code = kgr_vmcs_read32(VM_EXIT_INTR_ERROR_CODE);

	/*
	 * The #PF with PFEC.RSVD = 1 indicates the guest is accessing
	 * MMIO, it is better to report an internal error.
	 * See the comments in vmx_handle_exit.
	 */
	if ((vect_info & VECTORING_INFO_VALID_MASK) &&
	    !(kgr_is_page_fault(intr_info) && !(error_code & PFERR_RSVD_MASK))) {
		vcpu->run->exit_reason = KVM_EXIT_INTERNAL_ERROR;
		vcpu->run->internal.suberror = KVM_INTERNAL_ERROR_SIMUL_EX;
		vcpu->run->internal.ndata = 3;
		vcpu->run->internal.data[0] = vect_info;
		vcpu->run->internal.data[1] = intr_info;
		vcpu->run->internal.data[2] = error_code;
		return 0;
	}

	if (kgr_is_page_fault(intr_info)) {
		/* EPT won't cause page fault directly */
		BUG_ON(*kgr_enable_ept);
		cr2 = kgr_vmcs_readl(EXIT_QUALIFICATION);
		kgr_trace_kvm_page_fault(cr2, error_code);

		if (kgr_kvm_event_needs_reinjection(vcpu))
			kgr_kvm_mmu_unprotect_page_virt(vcpu, cr2);
		return kgr_kvm_mmu_page_fault(vcpu, cr2, error_code, NULL, 0);
	}

	ex_no = intr_info & INTR_INFO_VECTOR_MASK;

	if (vmx->rmode.vm86_active && kgr_rmode_exception(vcpu, ex_no))
		return kgr_handle_rmode_exception(vcpu, ex_no, error_code);

	switch (ex_no) {
	case AC_VECTOR:
		kgr_kvm_queue_exception_e(vcpu, AC_VECTOR, error_code);
		return 1;
	case DB_VECTOR:
		dr6 = kgr_vmcs_readl(EXIT_QUALIFICATION);
		if (!(vcpu->guest_debug &
		      (KVM_GUESTDBG_SINGLESTEP | KVM_GUESTDBG_USE_HW_BP))) {
			vcpu->arch.dr6 &= ~15;
			vcpu->arch.dr6 |= dr6 | DR6_RTM;
			/*
			 * Fix CVE-2018-1087
			 *  -1 line, +1 line
			 */
			if (kgr_is_icebp(intr_info))
				kgr_skip_emulated_instruction(vcpu);

			kgr_kvm_queue_exception(vcpu, DB_VECTOR);
			return 1;
		}
		kvm_run->debug.arch.dr6 = dr6 | DR6_FIXED_1;
		kvm_run->debug.arch.dr7 = kgr_vmcs_readl(GUEST_DR7);
		/* fall through */
	case BP_VECTOR:
		/*
		 * Update instruction length as we may reinject #BP from
		 * user space while in guest debugging mode. Reading it for
		 * #DB as well causes no harm, it is not used in that case.
		 */
		vmx->vcpu.arch.event_exit_inst_len =
			kgr_vmcs_read32(VM_EXIT_INSTRUCTION_LEN);
		kvm_run->exit_reason = KVM_EXIT_DEBUG;
		rip = kgr_kvm_rip_read(vcpu);
		kvm_run->debug.arch.pc = kgr_vmcs_readl(GUEST_CS_BASE) + rip;
		kvm_run->debug.arch.exception = ex_no;
		break;
	default:
		kvm_run->exit_reason = KVM_EXIT_EXCEPTION;
		kvm_run->ex.exception = ex_no;
		kvm_run->ex.error_code = error_code;
		break;
	}
	return 0;
}



static int kgr_patch_bsc1090869_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1090869_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1090869_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
  }

static struct notifier_block kgr_patch_bsc1090869_module_nb = {
	.notifier_call = kgr_patch_bsc1090869_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1090869_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1090869_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1090869_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1090869_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1090869_module_nb);
}

#endif /* IS_ENABLED(CONFIG_KVM_INTEL) */
